import { beforeEach, describe, expect, it, jest } from '@jest/globals';
import { execZNoOp } from '@run-z/exec-z';
import { TestPlan } from '../../../spec/test-plan.js';
import { ZJob } from '../../jobs/job.js';
import { ZShell } from '../../jobs/shell.js';
import { ZTaskParams } from '../../plan/task-params.js';
import { ScriptZTask } from './script.task.js';

describe('ScriptZTask', () => {
  let testPlan: TestPlan;

  beforeEach(() => {
    testPlan = new TestPlan();
  });

  it('does not contain any parameters', async () => {
    testPlan.addPackage('test', {
      packageJson: {
        scripts: {
          test: 'exec --arg',
        },
      },
    });

    const call = await testPlan.call('test');

    expect(call.task).toBeInstanceOf(ScriptZTask);

    const params = call.params(ZTaskParams.newEvaluator());

    expect(params.attrs).toEqual({});
    expect(params.args).toHaveLength(0);
  });

  describe('exec', () => {
    it('executes NPM script', async () => {
      const shell = {
        execScript: jest.fn<ZShell['execScript']>(execZNoOp),
      };

      testPlan.addPackage('test', {
        packageJson: {
          scripts: {
            test: 'run-z exec/--arg1 --arg2',
            exec: 'start --arg3',
          },
        },
      });

      const call = await testPlan.call('test');

      await call.exec(shell as Partial<ZShell> as ZShell).whenDone();

      expect(shell.execScript).toHaveBeenCalledWith(
        expect.objectContaining({
          call: await testPlan.callOf(call.task.target, 'exec'),
        }) as unknown as ZJob,
        'exec',
      );

      const { params } = shell.execScript.mock.calls[0][0];

      expect(params.args).toEqual(['--arg1', '--arg2']);
    });

    it('applies `+cmd:command` annex parameters', async () => {
      const shell = {
        execScript: jest.fn<ZShell['execScript']>(execZNoOp),
      };

      testPlan.addPackage('test', {
        packageJson: {
          scripts: {
            test: 'run-z exec/--arg1 --arg2',
            exec: 'start --arg3',
          },
        },
      });

      const call = await testPlan.parse('run-z test +cmd:start/--cmd-arg/cmd=applied');

      await call.exec(shell as Partial<ZShell> as ZShell).whenDone();

      expect(shell.execScript).toHaveBeenCalledWith(
        expect.objectContaining({
          call: await testPlan.callOf(call.task.target, 'exec'),
        }) as unknown as ZJob,
        'exec',
      );

      const { params } = shell.execScript.mock.calls[0][0];

      expect(params.args).toEqual(['--cmd-arg', '--arg1', '--arg2']);
      expect(params.attrs).toEqual({
        cmd: ['applied'],
      });
    });

    it('does not apply `+cmd:command` annex parameters to non-parsable command', async () => {
      const shell = {
        execScript: jest.fn<ZShell['execScript']>(execZNoOp),
      };

      testPlan.addPackage('test', {
        packageJson: {
          scripts: {
            test: 'run-z exec/--arg1 --arg2',
            exec: 'start --arg3 > out',
          },
        },
      });

      const call = await testPlan.parse('run-z test +cmd:start/--cmd-arg/cmd=applied');

      await call.exec(shell as Partial<ZShell> as ZShell).whenDone();

      expect(shell.execScript).toHaveBeenCalledWith(
        expect.objectContaining({
          call: await testPlan.callOf(call.task.target, 'exec'),
        }) as unknown as ZJob,
        'exec',
      );

      const { params } = shell.execScript.mock.calls[0][0];

      expect(params.args).toEqual(['--arg1', '--arg2']);
      expect(params.attrs).toEqual({});
    });
  });
});
